"""REST client handling, including MessagebirdStream base class."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Iterable, Optional
from urllib.parse import parse_qsl, urlparse

import requests
from singer_sdk.helpers.jsonpath import extract_jsonpath
from singer_sdk.pagination import BaseHATEOASPaginator, BaseOffsetPaginator
from singer_sdk.streams import RESTStream

SCHEMAS_DIR = Path(__file__).parent / Path("./schemas")


class MessagebirdHATEOASPaginator(BaseHATEOASPaginator):
    def get_next_url(self, response):
        return response.json()["links"]["next"]


class MessagebirdOffsetPaginator(BaseOffsetPaginator):
    def get_next(self, response) -> int | None:
        response_json = response.json()
        return response_json["offset"] + response_json["limit"]

    def has_more(self, response) -> bool:
        response_json = response.json()
        offset = response_json["offset"]
        count = response_json["count"]
        total_count = response_json["totalCount"]
        return count + offset <= total_count


class MessagebirdStream(RESTStream):
    """Messagebird stream class."""

    url_base = "https://rest.messagebird.com"

    records_jsonpath = "$.items[*]"
    next_page_token_jsonpath = "$.next_page"  # Or override `get_next_page_token`.
    _LOG_REQUEST_METRIC_URLS: bool = True

    def get_new_paginator(self):
        return MessagebirdHATEOASPaginator()

    @property
    def http_headers(self) -> dict:
        """Return the http headers needed."""
        headers = {}
        if "user_agent" in self.config:
            headers["User-Agent"] = self.config.get("user_agent")
        headers["Authorization"] = f"AccessKey {self.config['api_key']}"
        return headers

    def get_url_params(
        self, context: Optional[dict], next_page_token: Optional[Any]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization."""
        params: dict = {}
        if next_page_token:
            return dict(parse_qsl(next_page_token.query))
        if self.replication_key:
            params["sort"] = "asc"
            params["order_by"] = self.replication_key
        return params

    def parse_response(self, response: requests.Response) -> Iterable[dict]:
        """Parse the response and return an iterator of result records."""
        yield from extract_jsonpath(self.records_jsonpath, input=response.json())

    def response_error_message(self, response: requests.Response) -> str:
        """Build error message for invalid http statuses.

        WARNING - Override this method when the URL path may contain secrets or PII

        Args:
            response: A `requests.Response`_ object.

        Returns:
            str: The error message
        """
        full_path = urlparse(response.url).path or self.path
        error_content = ""
        if 400 <= response.status_code < 500:
            error_type = "Client"
            error_content = response.json()
        else:
            error_type = "Server"

        return (
            f"{response.status_code} {error_type} Error: "
            f"{response.reason} for path: {full_path} . "
            f"{error_content=}"
        )



"""Stream type classes for tap-messagebird."""

from pathlib import Path
from typing import Any, Dict, Optional

from tap_messagebird.client import MessagebirdOffsetPaginator, MessagebirdStream

SCHEMAS_DIR = Path(__file__).parent / Path("./schemas")


class MessagebirdConversations(MessagebirdStream):
    url_base = "https://conversations.messagebird.com/v1"

    def limit(self):
        return 20

    def get_new_paginator(self):
        return MessagebirdOffsetPaginator(
            start_value=0, page_size=self.limit()
        )  # type: ignore

    def get_url_params(
        self, context: Optional[dict], next_page_token: Optional[Any]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization.
        Overrode as we have a different paginator for this api
        """
        params = {}
        if next_page_token:
            params["offset"] = next_page_token
        params["status"] = "all"
        params["limit"] = self.limit()
        return params


class ConversationsStream(MessagebirdConversations):
    """Conversations stream."""

    name = "conversation"
    path = "/conversations"
    primary_keys = ["id"]
    replication_key = None
    # Optionally, you may also use `schema_filepath` in place of `schema`:
    schema_filepath = SCHEMAS_DIR / "conversation.json"

    def get_child_context(self, record: dict, context: Optional[dict]) -> dict:
        """Return a context dictionary for child streams."""
        return {
            "_sdc_conversations_id": record["id"],
        }


class ConversationMessagesStream(MessagebirdConversations):
    """Conversation Messages stream. Messages stream doesn't pull all messages as we were expecting"""

    name = "conversation_message"
    path = "/conversations/{_sdc_conversations_id}/messages"
    primary_keys = ["id"]
    replication_key = None
    # Optionally, you may also use `schema_filepath` in place of `schema`:
    schema_filepath = SCHEMAS_DIR / "conversation_message.json"
    parent_stream_type = ConversationsStream

    def get_url_params(
        self, context: Optional[dict], next_page_token: Optional[Any]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization."""
        params = super().get_url_params(
            context=context, next_page_token=next_page_token
        )
        if params.get("from") is None:
            params["from"] = self.config["start_date"]
        return params


class MessagesStream(MessagebirdStream):
    """Messages stream."""

    name = "message"
    path = "/messages"
    primary_keys = ["id"]
    replication_key = None
    # Optionally, you may also use `schema_filepath` in place of `schema`:
    schema_filepath = SCHEMAS_DIR / "message.json"

    def get_url_params(
        self, context: Optional[dict], next_page_token: Optional[Any]
    ) -> Dict[str, Any]:
        """Return a dictionary of values to be used in URL parameterization."""
        params = super().get_url_params(
            context=context, next_page_token=next_page_token
        )
        if params.get("from") is None:
            params["from"] = self.config["start_date"]
        return params



"""Messagebird tap class."""

from typing import List

import pendulum
from singer_sdk import Stream, Tap
from singer_sdk import typing as th  # JSON schema typing helpers

from tap_messagebird.streams import (
    ConversationMessagesStream,
    ConversationsStream,
    MessagesStream,
)

STREAM_TYPES = [
    ConversationsStream,
    MessagesStream,
    ConversationMessagesStream,
]


class TapMessagebird(Tap):
    """Messagebird tap class."""

    name = "tap-messagebird"

    config_jsonschema = th.PropertiesList(
        th.Property(
            "api_key",
            th.StringType,
            required=True,
            secret=True,
            description=(
                "The token to authenticate against the API service. "
                "Test keys are not supported for Conversations see "
                "https://support.messagebird.com/hc/en-us/articles/360000670709-What-is-the-difference-between-a-live-key-and-a-test-key-"  # noqa: E501
            ),
        ),
        th.Property(
            "start_date",
            th.DateTimeType,
            default=pendulum.now().subtract(years=3).set(tz="UTC").to_iso8601_string(),
            description=(
                "When to pull records starting at what date. "
                "ISO8601 format of date, defaults to 3 years ago."
            ),
        ),
    ).to_dict()

    def discover_streams(self) -> List[Stream]:
        """Return a list of discovered streams."""
        return [stream_class(tap=self) for stream_class in STREAM_TYPES]


if __name__ == "__main__":
    TapMessagebird.cli()



"""Tests standard tap features using the built-in SDK tests library."""

import os

from singer_sdk.testing import get_standard_tap_tests

from tap_messagebird.tap import TapMessagebird

SAMPLE_CONFIG = {
    "api_key": os.getenv("TAP_MESSAGEBIRD_API_KEY"),
}


# Run standard built-in tap tests from the SDK:
def test_standard_tap_tests():
    """Run standard tap tests from the SDK."""
    tests = get_standard_tap_tests(TapMessagebird, config=SAMPLE_CONFIG)
    for test in tests:
        test()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tap_messagebird', 'tap_messagebird.tests']

package_data = \
{'': ['*'], 'tap_messagebird': ['schemas/*']}

install_requires = \
['requests>=2.25.1,<3.0.0', 'singer-sdk>=0.17.0,<0.18.0']

extras_require = \
{'s3': ['fs-s3fs>=1.1.1,<2.0.0']}

entry_points = \
{'console_scripts': ['tap-messagebird = '
                     'tap_messagebird.tap:TapMessagebird.cli']}

setup_kwargs = {
    'name': 'tap-messagebird',
    'version': '0.0.3',
    'description': '`tap-messagebird` is a Singer tap for Messagebird, built with the Meltano Singer SDK.',
    'long_description': "# `tap-messagebird`\n\nMessagebird tap class.\n\nBuilt with the [Meltano Singer SDK](https://sdk.meltano.com).\n\n## Capabilities\n\n* `catalog`\n* `discover`\n* `about`\n* `stream-maps`\n* `schema-flattening`\n\n## Settings\n\n| Setting             | Required | Default | Description |\n|:--------------------|:--------:|:-------:|:------------|\n| api_key             | True     | None    | The token to authenticate against the API service. Test keys are not supported for Conversations see https://support.messagebird.com/hc/en-us/articles/360000670709-What-is-the-difference-between-a-live-key-and-a-test-key- |\n| start_date          | False    | now -3 years | When to pull records starting at what date. ISO8601 format of date, defaults to 3 years ago. |\n| stream_maps         | False    | None    | Config object for stream maps capability. For more information check out [Stream Maps](https://sdk.meltano.com/en/latest/stream_maps.html). |\n| stream_map_config   | False    | None    | User-defined config values to be used within map expressions. |\n| flattening_enabled  | False    | None    | 'True' to enable schema flattening and automatically expand nested properties. |\n| flattening_max_depth| False    | None    | The max depth to flatten schemas. |\n\nA full list of supported settings and capabilities is available by running: `tap-messagebird --about`\n\n\n## Installation\n\nInstall from PyPi:\n\n```bash\npipx install tap-messagebird\n```\n\n\n### Configure using environment variables\n\nThis Singer tap will automatically import any environment variables within the working directory's\n`.env` if the `--config=ENV` is provided, such that config values will be considered if a matching\nenvironment variable is set either in the terminal context or in the `.env` file.\n\n### Source Authentication and Authorization\n\nNote that Conversations do not work without a Production API key\n\n## Usage\n\nYou can easily run `tap-messagebird` by itself or in a pipeline using [Meltano](https://meltano.com/).\n\n### Executing the Tap Directly\n\n```bash\ntap-messagebird --version\ntap-messagebird --help\ntap-messagebird --config CONFIG --discover > ./catalog.json\n```\n\n## Developer Resources\n\nFollow these instructions to contribute to this project.\n\n### Initialize your Development Environment\n\n```bash\npipx install poetry\npoetry install\n```\n\n### Create and Run Tests\n\nCreate tests within the `tap_messagebird/tests` subfolder and\n  then run:\n\n```bash\npoetry run pytest\n```\n\nYou can also test the `tap-messagebird` CLI interface directly using `poetry run`:\n\n```bash\npoetry run tap-messagebird --help\n```\n\n### Testing with [Meltano](https://www.meltano.com)\n\n_**Note:** This tap will work in any Singer environment and does not require Meltano.\nExamples here are for convenience and to streamline end-to-end orchestration scenarios._\n\nNext, install Meltano (if you haven't already) and any needed plugins:\n\n```bash\n# Install meltano\npipx install meltano\n# Initialize meltano within this directory\ncd tap-messagebird\nmeltano install\n```\n\nNow you can test and orchestrate using Meltano:\n\n```bash\n# Test invocation:\nmeltano invoke tap-messagebird --version\n# OR run a test `elt` pipeline:\nmeltano elt tap-messagebird target-jsonl\n```\n\n### SDK Dev Guide\n\nSee the [dev guide](https://sdk.meltano.com/en/latest/dev_guide.html) for more instructions on how to use the SDK to\ndevelop your own taps and targets.\n",
    'author': 'Meltano',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)
